/*                  Part of x2grbl
 *
 *  Copyright Johann Wilhelm <johann.wilhelm@9mal6.de> 2015
 *
 *  see Readme.md for detailed license and usage information!
*/
#include "qpainter2gcode.h"

#include <QDebug>
#include <QSettings>
#include <QVector2D>
#include <QPointF>
#include <QList>

#include <limits>

const double GCodePainterEngine::ps2mm=1/72.0;

QPainter2GCode::QPainter2GCode() : QPaintDevice()
{
    Engine=new GCodePainterEngine;
}

QPaintEngine *QPainter2GCode::paintEngine() const {
    return Engine;
}

QPainter2GCode::~QPainter2GCode() {
    delete Engine;
}

void QPainter2GCode::InitParser(GCodeParser &Parser) {
    Parser.Lines.clear();
    const int count=Engine->GCodeListing.count();
    for (int i=0;i<count;i++) {
        Parser.Lines.insert(i+1,Engine->GCodeListing[i].serialize());
    }
}




GCodePainterEngine::GCodePainterEngine(PaintEngineFeatures features/*=0*/) : QPaintEngine(features) {
}

GCodePainterEngine::~GCodePainterEngine() {
}


bool GCodePainterEngine::begin(QPaintDevice *pdev) {
    highZ=1;
    lowZ=-.1;

    GCodeListing.clear();

    QSettings settings(QSettings::IniFormat, QSettings::UserScope, QString("2.5d Grbl-Commander"));

    settings.beginGroup("QPainter2GCode");
    if (settings.contains(QString("HighZ")))
        highZ=settings.value("HighZ").toDouble();
    if (settings.contains(QString("LowZ")))
        lowZ=settings.value("LowZ").toDouble();
    settings.endGroup();

    settings.beginGroup("MillingBits");
    QStringList Keys=settings.allKeys();
    for (int i=0;i<Keys.count();i++) {
        if (Keys[i].contains(QString("BitDiameter"))) {
            QString BitNumber=Keys[i];
            while (BitNumber.count() && !BitNumber[0].isDigit()) {
                BitNumber.remove(0,1);
            }

            int BitDiameter=settings.value(Keys[i]).toInt();
            Mills[BitNumber.toInt()]=BitDiameter;
        }
    }
    settings.endGroup();

    settings.beginGroup("DrillingBits");
    Keys=settings.allKeys();
    for (int i=0;i<Keys.count();i++) {
        if (Keys[i].contains(QString("BitDiameter"))) {
            QString BitNumber=Keys[i];
            while (BitNumber.count() && !BitNumber[0].isDigit()) {
                BitNumber.remove(0,1);
            }

            int BitDiameter=settings.value(Keys[i]).toInt();
            Drills[BitNumber.toInt()+1000]=BitDiameter;
        }
    }
    settings.endGroup();

    return true;
}

bool GCodePainterEngine::end() {
    return true;
}

void GCodePainterEngine::updateState(const QPaintEngineState &state) {

}

void GCodePainterEngine::drawRects(const QRect *rects, int rectCount) {
    QLine lineSegments[4];
    for (int i=0;i<rectCount;i++) {
        lineSegments[0].setP1(rects[i].topLeft());
        lineSegments[0].setP2(rects[i].topRight());
        lineSegments[1].setP1(rects[i].topRight());
        lineSegments[1].setP2(rects[i].bottomRight());
        lineSegments[2].setP1(rects[i].bottomRight());
        lineSegments[2].setP2(rects[i].bottomLeft());
        lineSegments[3].setP1(rects[i].bottomLeft());
        lineSegments[3].setP2(rects[i].topLeft());
        drawLines(lineSegments, 4);
    }
}

void GCodePainterEngine::drawRects(const QRectF *rects, int rectCount) {
    int width=(int)(painter()->pen().widthF()*10.0)*ps2mm;

    if (width==0)
        return;

    int millWidth=SetMillingTool(width);

    for (int i=0;i<rectCount;i++) {
        GCode Code;
        Code.Code=CodeType_GCode;

        Code.Cmd=0;
        Code.Parameters["Z"]=QString("%1").arg(highZ);
        GCodeListing.append(Code);

        Code.Parameters.clear();

        QPointF Point=rects[i].topLeft()*ps2mm;

        Code.Parameters["X"]=QString("%1").arg(Point.x());
        Code.Parameters["Y"]=QString("%1").arg(Point.y());
        GCodeListing.append(Code);

        Code.Parameters.clear();

        Code.Parameters["Z"]=QString("%1").arg(lowZ);
        GCodeListing.append(Code);

        Code.Parameters.clear();

        Point=rects[i].topRight()*ps2mm;

        Code.Cmd=1;
        Code.Parameters["X"]=QString("%1").arg(Point.x());
        Code.Parameters["Y"]=QString("%1").arg(Point.y());
        GCodeListing.append(Code);

        Point=rects[i].bottomRight()*ps2mm;

        Code.Cmd=1;
        Code.Parameters["X"]=QString("%1").arg(Point.x());
        Code.Parameters["Y"]=QString("%1").arg(Point.y());
        GCodeListing.append(Code);

        Point=rects[i].bottomLeft()*ps2mm;

        Code.Cmd=1;
        Code.Parameters["X"]=QString("%1").arg(Point.x());
        Code.Parameters["Y"]=QString("%1").arg(Point.y());
        GCodeListing.append(Code);

        Point=rects[i].topLeft()*ps2mm;

        Code.Cmd=1;
        Code.Parameters["X"]=QString("%1").arg(Point.x());
        Code.Parameters["Y"]=QString("%1").arg(Point.y());
        GCodeListing.append(Code);

        Code.Parameters.clear();

        Code.Cmd=0;
        Code.Parameters["Z"]=QString("%1").arg(highZ);
        GCodeListing.append(Code);

        Code.Parameters.clear();
    }
}

void GCodePainterEngine::drawLines(const QLine *lines, int lineCount) {
    qDebug()<<"GCodePainterEngine::drawLines(const QLine *lines, int lineCount) is unsupported!\n";
}

void GCodePainterEngine::drawLines(const QLineF *lines, int lineCount) {
    for (int i=0;i<lineCount;i++) {
        QPointF Start=lines[i].p1()*ps2mm;
        QPointF End=lines[i].p2()*ps2mm;

        int width=(int)(painter()->pen().widthF()*10.0)*ps2mm;
        int millWidth=SetMillingTool(width);

        GCode Code;
        Code.Cmd=0;
        Code.Code=CodeType_GCode;
        Code.Parameters["Z"]=QString("%1").arg(highZ);
        GCodeListing.append(Code);
        Code.Parameters.clear();

        Code.Parameters["X"]=QString("%1").arg(Start.x());
        Code.Parameters["Y"]=QString("%1").arg(Start.y());
        GCodeListing.append(Code);
        Code.Parameters.clear();

        Code.Parameters["Z"]=QString("%1").arg(lowZ);
        GCodeListing.append(Code);
        Code.Parameters.clear();

        Code.Cmd=1;
        Code.Parameters["X"]=QString("%1").arg(End.x());
        Code.Parameters["Y"]=QString("%1").arg(End.y());
        GCodeListing.append(Code);
        Code.Parameters.clear();

        Code.Cmd=0;
        Code.Parameters["Z"]=QString("%1").arg(highZ);
        GCodeListing.append(Code);
    }
}

void GCodePainterEngine::drawEllipse(const QRectF &r) {
    if (fabs(r.width()/r.height()-1)>0.001) { //1um roundnes to distuingish between circles and ellipse
        qDebug()<<"GCodePainterEngine::drawEllipse(const QRectF &r) only unsupports circles!\n";
    } else {
        int width=(int)( (painter()->pen().widthF()*ps2mm + r.width()*ps2mm) *10.0);

        bool CircleCanBeDrilled=false;
        if (SetDrillingTool(width)!= -1) {
            CircleCanBeDrilled=true;
        } else if (SetMillingTool(width) != -1) {
            CircleCanBeDrilled=true;
        }

        if (!CircleCanBeDrilled) {
            int width=(int)( (painter()->pen().widthF()*ps2mm) *10.0);
            int millWidth=SetMillingTool(width);
        }

        GCode Code;
        Code.Cmd=0;
        Code.Code=CodeType_GCode;
        Code.Parameters["Z"]=QString("%1").arg(highZ);
        GCodeListing.append(Code);
        Code.Parameters.clear();

        QVector2D Start;
        Start=QVector2D(r.center())-QVector2D(r.width()/2 + painter()->pen().width()/2,0);

        Code.Parameters["X"]=QString("%1").arg(Start.x());
        Code.Parameters["Y"]=QString("%1").arg(Start.y());
        GCodeListing.append(Code);
        Code.Parameters.clear();

        Code.Parameters["Z"]=QString("%1").arg(lowZ);
        GCodeListing.append(Code);
        Code.Parameters.clear();

        if (!CircleCanBeDrilled) {
            Code.Cmd=2;
            Code.Parameters["I"]=QString("%1").arg((r.width()-painter()->pen().width())/2);
            GCodeListing.append(Code);
            Code.Parameters.clear();
        }

        Code.Cmd=0;
        Code.Parameters["Z"]=QString("%1").arg(highZ);
        GCodeListing.append(Code);
    }
}

void GCodePainterEngine::drawEllipse(const QRect &r) {
    qDebug()<<"GCodePainterEngine::drawEllipse(const QRect &r) only unsupport!\n";
}

void GCodePainterEngine::drawPath(const QPainterPath &path) {
    bool bIsLow=false;

    GCode Code;
    Code.Cmd=0;
    Code.Code=CodeType_GCode;
    Code.Parameters["Z"]=QString("%1").arg(highZ);
    GCodeListing.append(Code);


    Code.Parameters.clear();

    if (painter()->brush().style()!=Qt::NoBrush) {
        QList<QPolygonF> FillPolygons = path.toFillPolygons(painter()->transform());

        for (int i=0;i<FillPolygons.count();i++) {

        }
    } else {
        QPointF currentPoint=QPointF(0,0);

        for (int elementIndex=0;elementIndex<path.elementCount();elementIndex++) {
             QPainterPath::Element element=path.elementAt(elementIndex);

             if (element.type==QPainterPath::MoveToElement) {
                 currentPoint=element;

                 Code.Parameters.clear();

                 if (bIsLow) {
                     Code.Cmd=0;
                     Code.Parameters["Z"]=QString("%1").arg(highZ);
                     GCodeListing.append(Code);

                     bIsLow=false;
                     Code.Parameters.clear();
                 }

                 Code.Parameters["X"]=QString("%1").arg(element.x*ps2mm);
                 Code.Parameters["Y"]=QString("%1").arg(element.y*ps2mm);
                 GCodeListing.append(Code);

             } else if (element.type==QPainterPath::LineToElement) {
                 currentPoint=element;

                 Code.Parameters.clear();

                 if (!bIsLow) {
                     Code.Cmd=0;
                     Code.Parameters["Z"]=QString("%1").arg(lowZ);
                     GCodeListing.append(Code);

                     bIsLow=true;
                     Code.Parameters.clear();
                 }

                 Code.Cmd=1;
                 Code.Parameters["X"]=QString("%1").arg(element.x*ps2mm);
                 Code.Parameters["Y"]=QString("%1").arg(element.y*ps2mm);
                 GCodeListing.append(Code);


             } else if (element.type==QPainterPath::CurveToElement) {
                 Code.Parameters.clear();

                 if (!bIsLow) {
                     Code.Cmd=0;
                     Code.Parameters["Z"]=QString("%1").arg(lowZ);
                     GCodeListing.append(Code);

                     bIsLow=true;
                     Code.Parameters.clear();
                 }

                 QVector2D BezierFrom=QVector2D(currentPoint);

                 QList<QVector2D> ControlPoints;

                 ControlPoints.append(QVector2D(element));

                 elementIndex++;
                 while (elementIndex<path.elementCount()) {
                     if (path.elementAt(elementIndex).type!=QPainterPath::CurveToDataElement)
                         break;
                     ControlPoints.append((QVector2D)(QPointF)path.elementAt(elementIndex));
                     elementIndex++;
                 }
                 elementIndex--;


                 QVector2D BezierTo=QVector2D(ControlPoints.last());
                 currentPoint=BezierTo.toPointF();

                 ControlPoints.removeLast();

                 if (ControlPoints.count()==2) {
                     Code.Cmd=1;

                     const int NumParts=10;
                     for(int i=0;i<NumParts;i++) {
                        const double t=(double)i/(double)NumParts;
                        double c1=pow(1-t,3);
                        double c2=3*t*pow(1-t,2);
                        double c3=3*pow(t,2)*(1-t);
                        double c4=pow(t,3);
                        QVector2D Point=BezierFrom*c1+ControlPoints[0]*c2+ControlPoints[1]*c3+BezierTo*c4;

                        Code.Parameters["X"]=QString("%1").arg(Point.x()*ps2mm);
                        Code.Parameters["Y"]=QString("%1").arg(Point.y()*ps2mm);
                        GCodeListing.append(Code);

                     }

                     Code.Parameters["X"]=QString("%1").arg(BezierTo.x()*ps2mm);
                     Code.Parameters["Y"]=QString("%1").arg(BezierTo.y()*ps2mm);
                     GCodeListing.append(Code);

                     Code.Parameters.clear();

                 } else if (ControlPoints.count()==1) {
                     Code.Cmd=1;

                     const int NumParts=2;
                     for(int i=0;i<NumParts;i++) {
                        const double t=(double)i/(double)NumParts;
                        QVector2D Point=BezierFrom*pow(1-t,2)+ControlPoints[0]*2*t*(1-t)+BezierTo*pow(t,2);

                        Code.Parameters["X"]=QString("%1").arg(Point.x()*ps2mm);
                        Code.Parameters["Y"]=QString("%1").arg(Point.y()*ps2mm);
                        GCodeListing.append(Code);

                     }

                     Code.Parameters["X"]=QString("%1").arg(BezierTo.x()*ps2mm);
                     Code.Parameters["Y"]=QString("%1").arg(BezierTo.y()*ps2mm);
                     GCodeListing.append(Code);
                     Code.Parameters.clear();
                 } else {
                     Code.Cmd=1;
                     Code.Parameters["X"]=QString("%1").arg(BezierTo.x()*ps2mm);
                     Code.Parameters["Y"]=QString("%1").arg(BezierTo.y()*ps2mm);
                     GCodeListing.append(Code);
                     Code.Parameters.clear();
                 }
             }

        }

        Code.Cmd=0;
        Code.Parameters["Z"]=QString("%1").arg(highZ);
        GCodeListing.append(Code);
    }

  //  qDebug()<<"GCodePainterEngine::drawPath(const QPainterPath &path) is unsupported!\n";
}

void GCodePainterEngine::drawPoints(const QPointF *points, int pointCount) {
    qDebug()<<"GCodePainterEngine::drawPoints(const QPointF *points, int pointCount) is unsupported!\n";
}

void GCodePainterEngine::drawPoints(const QPoint *points, int pointCount) {
    qDebug()<<"GCodePainterEngine::drawPoints(const QPoint *points, int pointCount) is unsupported!\n";
}

void GCodePainterEngine::drawPolygon(const QPointF *points, int pointCount, PolygonDrawMode mode) {
    qDebug()<<"GCodePainterEngine::drawPolygon(const QPointF *points, int pointCount, PolygonDrawMode mode) is unsupported!\n";
}

void GCodePainterEngine::drawPolygon(const QPoint *points, int pointCount, PolygonDrawMode mode) {
    qDebug()<<"GCodePainterEngine::drawPolygon(const QPoint *points, int pointCount, PolygonDrawMode mode) is unsupported!\n";
}

void GCodePainterEngine::drawPixmap(const QRectF &r, const QPixmap &pm, const QRectF &sr) {
    qDebug()<<"GCodePainterEngine::drawPixmap(const QRectF &r, const QPixmap &pm, const QRectF &sr) is unsupported!\n";
}

void GCodePainterEngine::drawTextItem(const QPointF &p, const QTextItem &textItem) {
    qDebug()<<"GCodePainterEngine::drawTextItem(const QPointF &p, const QTextItem &textItem) is unsupported!\n";
}

void GCodePainterEngine::drawTiledPixmap(const QRectF &r, const QPixmap &pixmap, const QPointF &s) {
    qDebug()<<"GCodePainterEngine::drawTiledPixmap(const QRectF &r, const QPixmap &pixmap, const QPointF &s) is unsupported!\n";
}

void GCodePainterEngine::drawImage(const QRectF &r, const QImage &pm, const QRectF &sr,
                           Qt::ImageConversionFlags flags /*= Qt::AutoColor*/) {
    qDebug()<<"GCodePainterEngine::drawImage(const QRectF &r, const QImage &pm, const QRectF &sr, Qt::ImageConversionFlags flags /*= Qt::AutoColor*/) is unsupported!\n";
}

QPaintEngine::Type GCodePainterEngine::type() const {
    return QPaintEngine::User;
}

int GCodePainterEngine::SetDrillingTool(int width) {
    QList<int>DrillIds=Drills.keys();

    int bestDrill=-1;
    double bestMatch=std::numeric_limits<bool>::max();
    bool perfectMatch=false;
    int DrillWidth;

    for (int i=0;i<DrillIds.count();i++) {
        if (width==Drills[DrillIds[i]]) {
            perfectMatch=true;
            bestDrill=DrillIds[i];
            DrillWidth=width;
            bestMatch=-1;
        }
    }

    if (!perfectMatch) {
        qDebug()<<"GCodePainterEngine only supports perfect line-width <-> tool-diameter matching!\n";
        return -1;
    }

    GCode Code;
    Code.Cmd=6;
    Code.Code=CodeType_MCode;
    Code.Parameters["T"]=QString("%1").arg(bestDrill);
    GCodeListing.append(Code);
    Code.Parameters.clear();

    return DrillWidth;
}

int GCodePainterEngine::SetMillingTool(int width) {
    QList<int>MillIds=Mills.keys();

    int bestMill=-1;
    double bestMatch=std::numeric_limits<bool>::max();
    bool perfectMatch=false;
    int MillWidth;

    for (int i=0;i<MillIds.count();i++) {
        if (width==Mills[MillIds[i]]) {
            perfectMatch=true;
            bestMill=MillIds[i];
            MillWidth=width;
            bestMatch=0;
        }
    }

    if (!perfectMatch) {
        qDebug()<<"GCodePainterEngine only supports perfect line-width <-> tool-diameter matching!\n";
        return -1;
    }

    GCode Code;
    Code.Cmd=6;
    Code.Code=CodeType_MCode;
    Code.Parameters["T"]=QString("%1").arg(bestMill);
    GCodeListing.append(Code);
    Code.Parameters.clear();

    return MillWidth;
}
